#include "Fibers.hpp"
#include <cassert>
#include <iostream>
#include <climits>

// Dummy function for fiber testing
void dummyFunction() {
    // Simple function for fiber execution
}

// Basic Fiber Creation and Properties
void testFiberCreation() {
    Fiber fiber(dummyFunction);
    assert(fiber.getState() == FiberState::CREATED);
    assert(fiber.getData() == nullptr);
    assert(fiber.getPriority() == 0);  // Default priority
    std::cout << "✓ Basic fiber creation test passed" << std::endl;
}

// State Transition Testing
void testFiberStateTransitions() {
    Fiber fiber(dummyFunction);
    
    // Test valid state transitions
    fiber.setState(FiberState::READY);
    assert(fiber.getState() == FiberState::READY);
    
    fiber.setState(FiberState::RUNNING);
    assert(fiber.getState() == FiberState::RUNNING);
    
    fiber.setState(FiberState::SUSPENDED);
    assert(fiber.getState() == FiberState::SUSPENDED);
    
    fiber.setState(FiberState::TERMINATED);
    assert(fiber.getState() == FiberState::TERMINATED);
    
    std::cout << "✓ State transitions test passed" << std::endl;
}

// Priority Management Testing
void testFiberPriority() {
    // Test different priority levels
    Fiber fiber1(dummyFunction, nullptr, 5);
    Fiber fiber2(dummyFunction, nullptr, 10);
    
    assert(fiber1.getPriority() == 5);
    assert(fiber2.getPriority() == 10);
    
    // Test priority modification
    fiber1.setPriority(8);
    assert(fiber1.getPriority() == 8);
    
    std::cout << "✓ Priority management test passed" << std::endl;
}

// Memory Allocation Testing
void testFiberMemoryAllocation() {
    try {
        Fiber fiber(dummyFunction);
        std::cout << "✓ Memory allocation test passed" << std::endl;
    } catch (const std::bad_alloc &e) {
        std::cerr << "✗ Memory allocation test failed: " << e.what() << std::endl;
        throw; // Re-throw to indicate test failure
    }
}

// Data Management Testing
void testDataHandling() {
    int testData = 42;
    Fiber fiber(dummyFunction, &testData);
    
    // Test data retrieval
    assert(*(int*)fiber.getData() == 42);
    
    // Test null data
    Fiber nullFiber(dummyFunction);
    assert(nullFiber.getData() == nullptr);
    
    std::cout << "✓ Data handling test passed" << std::endl;
}

// Multiple Fibers Testing
void testMultipleFibers() {
    Fiber fiber1(dummyFunction, nullptr, 5);
    Fiber fiber2(dummyFunction, nullptr, 3);
    Fiber fiber3(dummyFunction, nullptr, 7);
    
    // Test relative priorities
    assert(fiber3.getPriority() > fiber1.getPriority());
    assert(fiber1.getPriority() > fiber2.getPriority());
    
    std::cout << "✓ Multiple fibers test passed" << std::endl;
}

// State Logging Test
void testStateLogging() {
    Fiber fiber(dummyFunction);
    std::cout << "\nTesting State Logging:" << std::endl;
    
    fiber.setState(FiberState::READY);
    fiber.setState(FiberState::RUNNING);
    fiber.setState(FiberState::SUSPENDED);
    fiber.setState(FiberState::TERMINATED);
    
    std::cout << "✓ State logging test completed" << std::endl;
}

int main() {
   // Header
   std::cout << "\n========== Fiber Unit Tests ==========\n" << std::endl;
   
   try {
       // Basic Tests
       std::cout << "Running Basic Tests...\n" << std::endl;
       testFiberCreation();
       std::cout << "--------------------------------------------\n" << std::endl;

       testFiberStateTransitions(); 
       std::cout << "--------------------------------------------\n" << std::endl;

       // Priority Tests
       std::cout << "Running Priority Tests...\n" << std::endl;
       testFiberPriority();
       std::cout << "--------------------------------------------\n" << std::endl;

       // Memory Tests
       std::cout << "Running Memory Tests...\n" << std::endl;
       testFiberMemoryAllocation();
       std::cout << "--------------------------------------------\n" << std::endl;

       // Data Tests
       std::cout << "Running Data Tests...\n" << std::endl;
       testDataHandling();
       std::cout << "--------------------------------------------\n" << std::endl;

       // Multiple Fiber Tests
       std::cout << "Running Multiple Fiber Tests...\n" << std::endl;
       testMultipleFibers();
       std::cout << "--------------------------------------------\n" << std::endl;

       // State Tests
       std::cout << "Running State Tests...\n" << std::endl;
       testStateLogging();
       std::cout << "--------------------------------------------\n" << std::endl;

       // Success Message
       std::cout << "\n✓ All tests completed successfully!\n" << std::endl;
       std::cout << "=========================================\n" << std::endl;

   } catch (const std::exception &e) {
       std::cerr << "Test Suite Failed: " << e.what() << std::endl;
       return 1;
   }
   
   return 0;
}