#include <iostream>
#include <deque>
#include "../context/context.hpp" // Assuming the provided context library is included here

// Fiber class definition
class Fiber {
private:
    Context context_{}; // Holds the context of the fiber
    char *stack_bottom; // Stack memory for the fiber
    char *stack_top; // Stack memory for the fiber

public:
    Fiber(void (*function)()){
        // Allocate and initialize the stack for the fiber
        stack_bottom = new char[4096];
        stack_top = stack_bottom + 4096;
        uintptr_t sp = reinterpret_cast<uintptr_t>(stack_top);
        sp = (sp & -16L) - 128; // Align stack to 16 bytes and account for red zone
        context_.rsp = reinterpret_cast<void *>(sp);
        context_.rip = reinterpret_cast<void *>(function);
    }

    ~Fiber() {
        delete[] stack_bottom;
    }

    Context *getContext() { return &context_; }
};

// Scheduler class definition
class Scheduler {
private:
    std::deque<Fiber *> fibers_; // Queue of fibers to execute
    Context context_; // Context of the scheduler itself

public:
    Scheduler() {}

    ~Scheduler() {}

    void spawn(Fiber *fiber) {
        fibers_.push_back(fiber);
    }

    void do_it() {
        get_context(&context_);
        while (!fibers_.empty()) {
            Fiber *current_ = fibers_.front();
            fibers_.pop_front();

            // Switch to the fiber's context
            set_context(current_->getContext());
        }
    }

    void fiber_exit() {
        // Switch back to the scheduler's context
        set_context(&context_);
    }
};

// Global scheduler
Scheduler scheduler;

// Example functions for fibers
void func1() {
    std::cout << "fiber 1\n";
    scheduler.fiber_exit();
}

void func2() {
    std::cout << "fiber 2\n";
    scheduler.fiber_exit();
}

void func3() {
    std::cout << "fiber 3\n";
    scheduler.fiber_exit();
}

void func4() {
    std::cout << "fiber 4\n";
    scheduler.fiber_exit();
}

void child_function() {
    std::cout << "Child fiber\n";
    scheduler.fiber_exit();
}

void parent_function() {
    std::cout << "Parent fiber\n";
    // Create a child fiber inside the child function
    Fiber *child_fiber = new Fiber(child_function);

    // Spawn the parent fiber
    scheduler.spawn(child_fiber);

    scheduler.fiber_exit();
}


// Test functions
void testSchedulerBasic() {
    std::cout << "Running Basic Test: Scheduler with 2 fibers\n";
    
    Fiber f1(func1);
    Fiber f2(func2);

    scheduler.spawn(&f1);
    scheduler.spawn(&f2);

    scheduler.do_it(); // Run all fibers
}

void testSchedulerMultipleFibers() {
    std::cout << "Running Multiple Fibers Test\n";
    
    Fiber f1(func1);
    Fiber f2(func2);
    Fiber f3(func3);
    Fiber f4(func4);

    scheduler.spawn(&f1);
    scheduler.spawn(&f2);
    scheduler.spawn(&f3);
    scheduler.spawn(&f4);

    scheduler.do_it(); // Run all fibers
}

void testParentWithChildFiber() {
    std::cout << "Running Parent-Child Fiber Test\n";
    
    Fiber f1(parent_function);

    scheduler.spawn(&f1);

    scheduler.do_it(); // Run all fibers, including the child fiber created inside the parent
}

void testSchedulerEmptyQueue() {
    std::cout << "Running Empty Queue Test\n";

    // Nothing to execute, so the program should not output anything
    scheduler.do_it();
}

// Main function to demonstrate the scheduler with test cases
int main() {
    testSchedulerBasic();
    std::cout << "-----------------------------\n";
    testSchedulerMultipleFibers();
    std::cout << "-----------------------------\n";
    testParentWithChildFiber();
    std::cout << "-----------------------------\n";
    testSchedulerEmptyQueue();

    return 0;
}
