#include <iostream>
#include <deque>
#include "../context/context.hpp" // Assuming the provided context library is included here

// Fiber class definition
class Fiber {
private:
    Context context_{}; // Holds the context of the fiber
    void *data_; // Pointer to data passed to the fiber
    char *stack_bottom; // Stack memory for the fiber
    char *stack_top; // Stack memory for the fiber

public:
    Fiber(void (*function)(), void *data = nullptr){
        // Allocate and initialize the stack for the fiber
        data_ = data;
        stack_bottom = new char[4096];
        stack_top = stack_bottom + 4096;
        uintptr_t sp = reinterpret_cast<uintptr_t>(stack_top);
        sp = (sp & -16L) - 128; // Align stack to 16 bytes and account for red zone
        context_.rsp = reinterpret_cast<void *>(sp);
        context_.rip = reinterpret_cast<void *>(function);
    }

    ~Fiber() {
        delete[] stack_bottom;
    }

    Context *getContext() { return &context_; }

    void *getData() const { return data_; }
};

// Scheduler class definition
class Scheduler {
private:
    std::deque<Fiber *> fibers_; // Queue of fibers to execute
    Context context_; // Context of the scheduler itself
    Fiber *current_fiber_; // Keep track of the currently running fiber

public:
    Scheduler(){
        current_fiber_ = nullptr;
    }

    ~Scheduler(){}

    void spawn(Fiber *fiber) {
        fibers_.push_back(fiber);
    }

    void do_it() {
        get_context(&context_);
        if (!fibers_.empty()) {
            Fiber *current = fibers_.front();
            fibers_.pop_front();
            current_fiber_ = current;  // Set the current fiber

            // Switch to the fiber's context
            set_context(current->getContext());

            // Control returns here when the fiber calls fiber_exit
            // At this point, current_fiber_ will hold the current fiber
        }
    }

    void fiber_exit() {
        // Switch back to the scheduler's context
        set_context(&context_);
    }

    // Function to access the data of the currently running fiber
    void *getCurrentFiberData() {
        if (current_fiber_) {
            return current_fiber_->getData();
        }
        return nullptr;
    }
};

// Global scheduler instance
Scheduler scheduler;


// Example functions
void func1() {
    void *dp = scheduler.getCurrentFiberData();
    if (dp) {
        std::cout << "fiber 1: " << *reinterpret_cast<int *>(dp) << std::endl;
        *reinterpret_cast<int *>(dp) += 1;  // Increment the data
    }
    scheduler.fiber_exit();
}

void func2() {
    void *dp = scheduler.getCurrentFiberData();
    if (dp) {
        std::cout << "fiber 2: " << *reinterpret_cast<int *>(dp) << std::endl;
    }
    scheduler.fiber_exit();
}

// Main function to demonstrate the scheduler
int main() {
    int d = 10;
    void *dp = &d; // Data pointer to be shared between fibers

    Fiber f1(func1, dp);  // func1 uses shared data
    Fiber f2(func2, dp);  // func2 uses shared data

    scheduler.spawn(&f1);
    scheduler.spawn(&f2);

    scheduler.do_it();

    return 0;
}
